# Do not edit this file directly. It has been autogenerated from
# litestar/contrib/sqlalchemy/repository/_async.py
from __future__ import annotations

from advanced_alchemy.repository import SQLAlchemySyncRepository

__all__ = ("SQLAlchemySyncRepository",)
